/**********************************************************
	expert.c
  
	Expert System Shell
	AI Project

        (11/13/1997,11/17/1997,11/19/1997)

	Copyright (c) 1997, Particle

        Compiles with no problems under Borland C++ v3.1,
        DJGPP v2 (protected mode dos compiler),
        Visual C++ v5 Pro, and UNIX "gcc."

***********************************************************/

/**********************************************************
	includes
	(11/13/1997)
***********************************************************/
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<ctype.h>

/**********************************************************
	type defines, structure defines, etc...
	(11/13/1997)
***********************************************************/

     /* The MAX_AX, MAX_RULES, and MAX_TERMS, can and should be made
        bigger, they're of this size to make them compilable under
        Borland C++ v3.1, if using a UNIX, or a Protected Mode
        compiler, set them to higher values */
#define MAX_AX 32
#define MAX_RULES 512
#define MAX_TERMS 1024

#define GENERAL_BUFFER_SIZE 1024

#define RULES_TOKEN "RULES"
#define GOALS_TOKEN "GOALS"
#define IF_TOKEN "IF"
#define THEN_TOKEN "THEN"

#define TRUE_RESPONSE "true"
#define FALSE_RESPONSE "false"
#define UNKNOWN_RESPONSE "unknown"

#ifndef DEBUG
	#define DEBUG 0
#endif

typedef enum {true, false, pending} boolean;

typedef struct rule_s{
	int ax_number;
	int	ax[MAX_AX];
	int conc;
}rule_t;

/**********************************************************
	some prototypes
        (11/17/1997,11/19/1997)
***********************************************************/
void print_rule(rule_t *);
void confirm(int);
boolean load_file(void);
boolean choose_execution_mode(void);
boolean read_in_data(FILE *);
void read_goal(char *);
void read_rule(char *);
void print_goals(void);
void print_rules(void);
void print_rule(rule_t *);
void print_strings(void);
void try_to_confirm(void);
void confirm(int);
boolean fires(rule_t *);
boolean findint(int *,int,int);
void init_stuff(void);
int findstring(char *);
boolean addstring(char *);

/**********************************************************
	globals
	(11/13/1997)
***********************************************************/
boolean do_trace;

int num_rules = 0;
rule_t rules[MAX_RULES];

int num_asked = 0;
int asked[MAX_TERMS];

int num_goals = 0;
int goals[MAX_TERMS];

	/*  contains an array of ALL things, with their proven,
		or unproven value, the pointer into this array, is
		the value of the number associated with the 
		"string"*/
boolean proved[MAX_TERMS];

int num_strings = 1;    /* can't start with 0 */
char *strings[MAX_TERMS];

/**********************************************************
function to add that string to a global list of currently used
strings... which are associated with an integer by a function
"find"
 (10/02/1997,11/13/1997)
***********************************************************/
boolean addstring(char *string)
{

#if DEBUG
	fprintf(stderr,"addstring(\"%s\");\n",string);
#endif

	if(num_strings >= MAX_TERMS)
		return false;
	strings[num_strings]=
			(char*)malloc((strlen(string)+1)*sizeof(char));
	if(!strings[num_strings])       return false;
	strcpy(strings[num_strings],string);
	num_strings++;
	return true;
}

/**********************************************************
	function which searches for a string in a global string 
	database. it returns -1 if not found, or index of string 
	if found...note: indexes start with 1, not 0.
	(10/02/1997,10/04/1997,11/13/1997)
***********************************************************/
int findstring(char *string)
{
	int i;

	/* has to start from 1, not 0 */
#if DEBUG
	fprintf(stderr,"findstring(\"%s\");\n",string);
#endif
	
	for(i=1;i<num_strings;i++)
		if(!strcmp(strings[i],string))
			return i;
	return -1;
}

/**********************************************************
	function to initialize stuff...
	(11/17/1997)
***********************************************************/
void init_stuff()
{
	int i;
	/* assume everything is askable */
	for(i=0;i<num_strings;i++)
		proved[i] = false;
	/* initialize everything to "false" (ie: unconfirmed)*/

}

/**********************************************************
	function to find an int in an int array
	(11/17/1997)
***********************************************************/
boolean findint(int *a,int n,int val)
{
	int i;
	for(i=0;i<n;i++){
		if(a[i] == val)
			return true;
	}
	return false;
}

/**********************************************************
	the function to fire rules
	(11/17/1997)
***********************************************************/
boolean fires(rule_t *r)
{
	int i;
	for(i=0;i<r->ax_number;i++){
		confirm(r->ax[i]);
		if(proved[r->ax[i]] != true)
			return false;
	}
	if(do_trace == true){
		fprintf(stdout,"Fired: ");
		print_rule(r);
	}
	return true;
}

/**********************************************************
	the confirm function, the heart of the system (almost)
	(11/17/1997)
***********************************************************/
void confirm(int h)
{	
	int i;
	char buffer[GENERAL_BUFFER_SIZE];
	if(proved[h] == true) return;
	if(proved[h] == pending) return;
	proved[h] = pending;
	for(i=0;i<num_rules;i++){
		if(rules[i].conc == h)
			if(fires(&rules[i]) == true){
				proved[h] = true;
				return;
			}
	}
	/* assume everything is askable, exept the goals 
		themselves, also make sure we don't ask 
		the same thing twice! */
	if(findint(goals,num_goals,h) == false &&
			findint(asked,num_asked,h) == false){
		asked[num_asked++] = h;
                fprintf(stdout,"is %s: true/false/unknown? ",strings[h]);
		fgets(buffer,GENERAL_BUFFER_SIZE,stdin);
              if(!strncmp(buffer,TRUE_RESPONSE,strlen(TRUE_RESPONSE)))
			proved[h] = true;          
              else proved[h] = false;
	}
	else proved[h] = false;
}

/**********************************************************
	function that goes through the global variables and 
	tries to confirm stuff...
	(11/13/1997,11/17/1997)
***********************************************************/
void try_to_confirm()
{
	int i;
#if DEBUG
	fprintf(stderr,"try_to_confirm();\n");
#endif
	for(i=0;i<num_goals;i++){
		confirm(goals[i]);
	}
	for(i=0;i<num_goals;i++){
		if(proved[goals[i]] == true)
			fprintf(stdout,"%s confirmed!\n",strings[goals[i]]);
		else fprintf(stdout,"%s didn't confirm!\n",strings[goals[i]]);
	}
}

/**********************************************************
	function to print the strings, the string values
	of numbers.
	(11/17/1997)
***********************************************************/
void print_strings()
{
	int i;
	for(i=1;i<num_strings;i++){
		fputs(strings[i],stdout);
		fputc('\n',stdout);
	}
}

/**********************************************************
	print 1 rule... (needed for that "output" option)
	(11/17/1997)
***********************************************************/
void print_rule(rule_t *r)
{
	int j;
	fputs(IF_TOKEN,stdout);
	fputc(' ',stdout);
	for(j=0;j<r->ax_number;j++){
		fputs(strings[r->ax[j]],stdout);
		fputc(' ',stdout);
	}
	fputs(THEN_TOKEN,stdout);
	fputc(' ',stdout);
	fputs(strings[r->conc],stdout);
	fputc('\n',stdout);
}

/**********************************************************
	function to print the rules, in their IF ... THEN ... 
	format.
	(11/17/1997)
***********************************************************/
void print_rules()
{
	int i,j;
	fputs(RULES_TOKEN,stdout);
	fputc('\n',stdout);
	for(i=0;i<num_rules;i++){
		fputs(IF_TOKEN,stdout);
		fputc(' ',stdout);
		for(j=0;j<rules[i].ax_number;j++){
			fputs(strings[rules[i].ax[j]],stdout);
			fputc(' ',stdout);
		}
		fputs(THEN_TOKEN,stdout);
		fputc(' ',stdout);
		fputs(strings[rules[i].conc],stdout);
		fputc('\n',stdout);
	}
}

/**********************************************************
	function to print goals, 
	(11/17/1997)
***********************************************************/
void print_goals()
{
	int i;
	fputs(GOALS_TOKEN,stdout);
	fputc('\n',stdout);
	for(i=0;i<num_goals;i++){
		fputs(strings[goals[i]],stdout);
		fputc(' ',stdout);
	}
	fputc('\n',stdout);
}

/**********************************************************
	function to read a rule, (parse it out...  ;-)
	(11/13/1997,11/17/1997)
***********************************************************/
void read_rule(char *buffer)
{
	int i,j;
	char temp[GENERAL_BUFFER_SIZE];
	enum {reading_ax,reading_cons} state;
	state = reading_ax;

	rules[num_rules].ax_number = 0;

#if DEBUG
	fprintf(stderr,"read_rule(\"%s\");\n",buffer);
#endif
		/* make sure a rule has "IF" in start */
	if(!strncmp(buffer,IF_TOKEN,i=strlen(IF_TOKEN))){
		while(buffer[i] != '\0'){
			while(isspace(buffer[i])) i++;
			j=0;
			while(!isspace(buffer[i]) && buffer[i]!='\0')
				temp[j++] = buffer[i++];
			temp[j] = '\0';

			if(!strncmp(temp,THEN_TOKEN,strlen(THEN_TOKEN))){
				state = reading_cons;
				continue;
			}
			if(findstring(temp) == -1)
				if(addstring(temp) == false){
					fprintf(stderr,"List Overflow!\n");
					exit(1);
				}
                                /* this time, it should be in */
                        j = findstring(temp);
			if(state == reading_ax)
                         rules[num_rules].ax[rules[num_rules].ax_number++]=j;
			else if(state == reading_cons){
				rules[num_rules].conc = j;
				break;
			}
		}
		num_rules++;
	}
}

/**********************************************************
	function to read a goal
	(11/13/1997,11/17/1997)
***********************************************************/
void read_goal(char *buffer)
{
	int i=0,j;
	char temp[GENERAL_BUFFER_SIZE];
	
#if DEBUG
	fprintf(stderr,"read_goal(\"%s\");\n",buffer);
#endif

	while(buffer[i] != '\0'){
		while(isspace(buffer[i])) i++;
		j=0;
		while(!isspace(buffer[i]) && buffer[i]!='\0')
			temp[j++] = buffer[i++];
			temp[j] = '\0';
		if(findstring(temp) == -1)
			if(addstring(temp) == false){
				fprintf(stderr,"List Overflow!\n");
				exit(1);
			}
		j = findstring(temp);	/* this time, it should be in */	
		goals[num_goals++] = j;
	}
}

/**********************************************************
	function to read the data from 'input' into a 
	global data structure
	(11/13/1997)
***********************************************************/
boolean read_in_data(FILE *input)
{
	char buffer[GENERAL_BUFFER_SIZE];
	enum {_goals,_rules} current;

	current=_rules;	/* default to reading rules */

#if DEBUG
	fprintf(stderr,"read_in_data(FILE*);\n");
#endif
	
	/* read the input one line at a time */
	while(fgets(buffer,GENERAL_BUFFER_SIZE,input)){
		buffer[strlen(buffer)-1] = '\0';
		if(!strcmp(buffer,RULES_TOKEN))
			current=_rules;
		else if(!strcmp(buffer,GOALS_TOKEN))
			current=_goals;
		else if(current == _rules)
			read_rule(buffer);
		else if(current == _goals)
			read_goal(buffer);
	}
	return true;
}

/**********************************************************
	function to choose execution mode, and set a 
	global variable.
	(11/13/1997)
***********************************************************/
boolean choose_execution_mode()
{
        char buffer[GENERAL_BUFFER_SIZE];

#if DEBUG
	fprintf(stderr,"choose_execution_mode();\n");
#endif

	fprintf(stderr,"Execute with trace of rule firing? y/n? ");
        if(!fgets(buffer,GENERAL_BUFFER_SIZE,stdin))
		return false;
	if(buffer[0] == 'Y' || buffer[0] == 'y')
		do_trace = true;
	else do_trace = false;
	return true;
}

/**********************************************************
	function to get file name from user, open a file,
	call a func to read data, and close the file
	(11/13/1997)
***********************************************************/
boolean load_file(){
	FILE *input;
        char buffer[GENERAL_BUFFER_SIZE];

#if DEBUG
	fprintf(stderr,"load_file();\n");
#endif

	fprintf(stderr,"Enter File: ");
        if(!fgets(buffer,GENERAL_BUFFER_SIZE,stdin))
		return false;
	buffer[strlen(buffer)-1] = '\0';
	input=fopen(buffer,"rt");
	if(input == NULL){
		fprintf(stderr,"Can't open: %s\n",buffer);
		return false;
	}
	if(read_in_data(input) == false){
		fprintf(stderr,"Can't read: %s\n",buffer);
		return false;
	}
	fclose(input);
	return true;
}


/**********************************************************
	the main
	(11/13/1997)
***********************************************************/
int main()
{
	if(load_file() == false){
		fprintf(stderr,"Was not able to load file.\n");
		return 1;
	}
		/* print what we'll be working with */
	print_rules();
	print_goals();

	if(choose_execution_mode() == false){
		fprintf(stderr,"Couldn't choose execution mode.\n");
		return 1;
	}
	init_stuff();

#if DEBUG
	print_strings();  */
	print_rules();
	print_goals();
#endif
		/* the function to do the rest of the work */
	try_to_confirm();
        return 0;
}

